﻿using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;

namespace everlaster
{
    sealed class SessionPlugin
    {
        readonly SessionPluginsHook _mainScript;
        readonly JSONStorable _storable;
        readonly Dictionary<JSONStorableBool, JSONStorableBool> _boolParams = new Dictionary<JSONStorableBool, JSONStorableBool>();
        readonly Dictionary<JSONStorableFloat, JSONStorableFloat> _floatParams = new Dictionary<JSONStorableFloat, JSONStorableFloat>();
        readonly Dictionary<JSONStorableString, JSONStorableString> _stringParams = new Dictionary<JSONStorableString, JSONStorableString>();
        readonly Dictionary<JSONStorableStringChooser, JSONStorableStringChooser> _stringChooserParams = new Dictionary<JSONStorableStringChooser, JSONStorableStringChooser>();
        readonly Dictionary<JSONStorableUrl, JSONStorableUrl> _urlParams = new Dictionary<JSONStorableUrl, JSONStorableUrl>();
        readonly Dictionary<JSONStorableColor, JSONStorableColor> _colorParams = new Dictionary<JSONStorableColor, JSONStorableColor>();
        readonly List<JSONStorableAction> _actions = new List<JSONStorableAction>();
        readonly List<JSONStorableActionAudioClip> _audioClipActions = new List<JSONStorableActionAudioClip>();

        public string Name { get; }
        public string FullName { get; }
        public bool IsRemoved => _storable == null;

        public SessionPlugin(string name, MVRScript script, SessionPluginsHook mainScript)
        {
            Name = name;
            FullName = BuildFullName(script);
            _storable = script;
            _mainScript = mainScript;
            RegisterParams();
        }

        string BuildFullName(MVRScript script)
        {
            var sb = new StringBuilder();
            sb.Append(Name);
            string packageId = script.GetPackageId();
            if(!string.IsNullOrEmpty(packageId))
            {
                string color = ColorUtility.ToHtmlStringRGB(Colors.darkGray);
                sb.AppendFormat(" <color=#{0}>[{1}.var]</color>", color, packageId);
            }

            return sb.ToString();
        }

        void RegisterParams()
        {
            var nameSb = new StringBuilder();

            foreach(string paramName in _storable.GetBoolParamNames())
            {
                var param = _storable.GetBoolJSONParam(paramName);
                var copy = new JSONStorableBool(BuildParamName(nameSb, paramName), param.defaultVal)
                {
                    val = param.val,
                };
                _boolParams[copy] = param;
                _boolParams[param] = copy;
                _mainScript.RegisterBool(copy);

                copy.setJSONCallbackFunction = OnBoolParamValueChanged;
                param.setJSONCallbackFunction += OnBoolParamValueChanged;
            }

            foreach(string paramName in _storable.GetFloatParamNames())
            {
                var param = _storable.GetFloatJSONParam(paramName);
                var copy = new JSONStorableFloat(BuildParamName(nameSb, paramName), param.defaultVal, param.min, param.max)
                {
                    val = param.val,
                };
                _floatParams[copy] = param;
                _floatParams[param] = copy;
                _mainScript.RegisterFloat(copy);

                copy.setJSONCallbackFunction = OnFloatParamValueChanged;
                param.setJSONCallbackFunction += OnFloatParamValueChanged;
            }

            foreach(string paramName in _storable.GetStringParamNames())
            {
                var param = _storable.GetStringJSONParam(paramName);
                var copy = new JSONStorableString(BuildParamName(nameSb, paramName), param.defaultVal)
                {
                    val = param.val,
                };
                _stringParams[copy] = param;
                _stringParams[param] = copy;
                _mainScript.RegisterString(copy);

                copy.setJSONCallbackFunction = OnStringParamValueChanged;
                param.setJSONCallbackFunction += OnStringParamValueChanged;
            }

            foreach(string paramName in _storable.GetStringChooserParamNames())
            {
                var param = _storable.GetStringChooserJSONParam(paramName);
                var copy = new JSONStorableStringChooser(BuildParamName(nameSb, paramName), param.choices, param.defaultVal, param.label)
                {
                    val = param.val,
                };
                _stringChooserParams[copy] = param;
                _stringChooserParams[param] = copy;
                _mainScript.RegisterStringChooser(copy);

                copy.setJSONCallbackFunction = OnStringChooserParamValueChanged;
                param.setJSONCallbackFunction += OnStringChooserParamValueChanged;
            }

            foreach(string paramName in _storable.GetUrlParamNames())
            {
                var param = _storable.GetUrlJSONParam(paramName);
                var copy = new JSONStorableUrl(BuildParamName(nameSb, paramName), param.defaultVal)
                {
                    val = param.val,
                };
                _urlParams[copy] = param;
                _urlParams[param] = copy;
                _mainScript.RegisterUrl(copy);

                copy.setJSONCallbackFunction = OnUrlParamValueChanged;
                param.setJSONCallbackFunction += OnUrlParamValueChanged;
            }

            foreach(string paramName in _storable.GetColorParamNames())
            {
                var param = _storable.GetColorJSONParam(paramName);
                var copy = new JSONStorableColor(BuildParamName(nameSb, paramName), param.defaultVal)
                {
                    val = param.val,
                };
                _colorParams[copy] = param;
                _colorParams[param] = copy;
                _mainScript.RegisterColor(copy);

                copy.setJSONCallbackFunction = OnColorParamValueChanged;
                param.setJSONCallbackFunction += OnColorParamValueChanged;
            }

            foreach(string paramName in _storable.GetActionNames().Except(_mainScript.BuiltInActionNames))
            {
                var action = _storable.GetAction(paramName);
                var copy = new JSONStorableAction(BuildParamName(nameSb, paramName), action.actionCallback);
                _mainScript.RegisterAction(copy);
                _actions.Add(copy);
            }

            foreach(string paramName in _storable.GetAudioClipActionNames())
            {
                var action = _storable.GetAudioClipAction(paramName);
                var copy = new JSONStorableActionAudioClip(BuildParamName(nameSb, paramName), action.actionCallback);
                _mainScript.RegisterAudioClipAction(copy);
                _audioClipActions.Add(copy);
            }
        }

        string BuildParamName(StringBuilder nameSb, string paramName) =>
            nameSb.Clear().AppendFormat("{0}_{1}", Name, paramName).ToString();

        void OnBoolParamValueChanged(JSONStorableBool jsb) => _boolParams[jsb]?.SetVal(jsb.val);
        void OnFloatParamValueChanged(JSONStorableFloat jsf) => _floatParams[jsf]?.SetVal(jsf.val);
        void OnStringParamValueChanged(JSONStorableString jss) => _stringParams[jss]?.SetVal(jss.val);
        void OnStringChooserParamValueChanged(JSONStorableStringChooser jssc) => _stringChooserParams[jssc]?.SetVal(jssc.val);

        void OnUrlParamValueChanged(JSONStorableString jss)
        {
            var jsUrl = (JSONStorableUrl) jss;
            _urlParams[jsUrl]?.SetVal(jss.val);
        }

        void OnColorParamValueChanged(JSONStorableColor jsc) => _colorParams[jsc]?.SetVal(jsc.val.H, jsc.val.S, jsc.val.V);

        public void Cleanup()
        {
            foreach(var pair in _boolParams)
            {
                pair.Value.setJSONCallbackFunction -= OnBoolParamValueChanged;
                _mainScript.DeregisterBool(pair.Key);
            }

            foreach(var pair in _floatParams)
            {
                pair.Value.setJSONCallbackFunction -= OnFloatParamValueChanged;
                _mainScript.DeregisterFloat(pair.Key);
            }

            foreach(var pair in _stringParams)
            {
                pair.Value.setJSONCallbackFunction -= OnStringParamValueChanged;
                _mainScript.DeregisterString(pair.Key);
            }

            foreach(var pair in _stringChooserParams)
            {
                pair.Value.setJSONCallbackFunction -= OnStringChooserParamValueChanged;
                _mainScript.DeregisterStringChooser(pair.Key);
            }

            foreach(var pair in _urlParams)
            {
                pair.Value.setJSONCallbackFunction -= OnUrlParamValueChanged;
                _mainScript.DeregisterUrl(pair.Key);
            }

            foreach(var pair in _colorParams)
            {
                pair.Value.setJSONCallbackFunction -= OnColorParamValueChanged;
                _mainScript.DeregisterColor(pair.Key);
            }

            foreach(var action in _actions)
            {
                _mainScript.DeregisterAction(action);
            }

            foreach(var action in _audioClipActions)
            {
                _mainScript.DeregisterAudioClipAction(action);
            }
        }
    }
}
