﻿//
using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using UnityEngine;
using UnityEngine.UI;

namespace everlaster
{
    class ScriptBase : MVRScript
    {
        // ReSharper disable once MemberCanBeProtected.Global
        public const string VERSION = "1.0.0";
        public bool? Initialized { get; set; }
        UnityEventsListener PluginUIEventsListener { get; set; }

        bool _uiOpened;

        // Prevent ScriptBase from showing up as a plugin in Plugins tab
        public override bool ShouldIgnore() => true;

        public override void InitUI()
        {
            if(ShouldIgnore())
            {
                return;
            }

            base.InitUI();
            if(!UITransform || PluginUIEventsListener)
            {
                return;
            }

            StartCoroutine(InitUICo());
        }

        protected virtual void OnUIEnabled()
        {
        }

        // ReSharper disable once VirtualMemberNeverOverridden.Global
        protected virtual void OnUIDisabled()
        {
        }

        IEnumerator InitUICo()
        {
            PluginUIEventsListener = UITransform.gameObject.AddComponent<UnityEventsListener>();
            PluginUIEventsListener.EnabledHandlers += () => StartCoroutine(OnUIEnabledCo());

            while(Initialized == null)
            {
                yield return null;
            }

            if(Initialized == false)
            {
                enabledJSON.val = false;
                yield break;
            }

            PluginUIEventsListener.DisabledHandlers += () => StartCoroutine(OnUIDisabledCo());
        }

        bool _inEnabledCo;

        IEnumerator OnUIEnabledCo()
        {
            if(_inEnabledCo)
            {
                /* When VAM UI is toggled back on with the plugin UI already active, onEnable gets called twice and onDisable once.
                 * This ensures onEnable logic executes just once.
                 */
                yield break;
            }

            _inEnabledCo = true;

            if(!_uiOpened)
            {
                SetGrayBackground();
            }

            while(Initialized == null)
            {
                yield return null;
            }

            if(Initialized == false)
            {
                yield break;
            }

            if(!_uiOpened)
            {
                _uiOpened = true;
            }

            yield return null;
            yield return null;
            yield return null;
            OnUIEnabled();

            #if ENV_DEVELOPMENT
            {
                if(!_devUICreated)
                {
                    CreateDevUI();
                }
            }
            #endif

            _inEnabledCo = false;
        }

        void SetGrayBackground()
        {
            var background = rightUIContent.parent.parent.parent.transform.GetComponent<Image>();
            background.color = Colors.backgroundGray;
        }

        IEnumerator OnUIDisabledCo()
        {
            if(_inEnabledCo)
            {
                /* When VAM UI is toggled back on with the plugin UI already active, onEnable gets called twice and onDisable once.
                 * This ensures only onEnable logic executes.
                 */
                yield break;
            }

            OnUIDisabled();
        }

        // ReSharper disable MemberCanBeProtected.Global
        // ReSharper disable MemberCanBePrivate.Global
        // ReSharper disable UnusedMember.Global

        public UIDynamicButton CreateCustomButton(string label, Vector2 position, Vector2 size, bool devConfigurable = true) =>
            CreateCustomButton(label, label, position, size, devConfigurable);

        public UIDynamicButton CreateCustomButton(string id, string label, Vector2 position, Vector2 size, bool configurable = true)
        {
            var t = this.InstantiateButton(UITransform);
            var rectT = Utils.GetRectTransform(t);
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicButton>();
            uiDynamic.label = label;

            #if ENV_DEVELOPMENT
            {
                if(configurable)
                {
                    AddRect(id, rectT);
                }
            }
            #endif

            return uiDynamic;
        }


        public UIDynamicTextField CreateCustomTextField(string text, Vector2 position, Vector2 size, bool devConfigurable = true) =>
            CreateCustomTextField(text, text, position, size, devConfigurable);

        public UIDynamicTextField CreateCustomTextField(string id, string text, Vector2 position, Vector2 size, bool configurable = true)
        {
            var t = this.InstantiateTextField(UITransform);
            var rectT = Utils.GetRectTransform(t);
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicTextField>();
            uiDynamic.text = text;

            #if ENV_DEVELOPMENT
            {
                if(configurable)
                {
                    AddRect(id.Substring(0, Math.Min(id.Length, 15)), rectT);
                }
            }
            #endif

            return uiDynamic;
        }

        public UIDynamicSlider CreateCustomSlider(string label, Vector2 position, bool devConfigurable = true) =>
            CreateCustomSlider(label, label, position, devConfigurable);

        public UIDynamicSlider CreateCustomSlider(string id, string label, Vector2 position, bool configurable = true)
        {
            var t = this.InstantiateSlider(UITransform);
            var rectT = Utils.GetRectTransform(t);
            rectT.anchoredPosition = position;
            rectT.sizeDelta = new Vector2(-555, 120);
            var uiDynamic = t.GetComponent<UIDynamicSlider>();
            uiDynamic.label = label;
            uiDynamic.rangeAdjustEnabled = false;

            #if ENV_DEVELOPMENT
            {
                if(configurable)
                {
                    AddRect(id, rectT);
                }
            }
            #endif

            return uiDynamic;
        }

        public UIDynamicToggle CreateCustomToggle(string label, Vector2 position, Vector2 size, bool devConfigurable = true) =>
            CreateCustomToggle(label, label, position, size, devConfigurable);

        public UIDynamicToggle CreateCustomToggle(string id, string label, Vector2 position, Vector2 size, bool configurable = true)
        {
            var t = this.InstantiateToggle(UITransform);
            var rectT = Utils.GetRectTransform(t);
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicToggle>();
            uiDynamic.label = label;

            #if ENV_DEVELOPMENT
            {
                if(configurable)
                {
                    AddRect(id, rectT);
                }
            }
            #endif

            return uiDynamic;
        }

        public UIDynamicPopup CreateCustomPopup(string label, Vector2 position, Vector2 size, bool devConfigurable = true) =>
            CreateCustomFilterablePopup(label, label, position, size, devConfigurable);

        public UIDynamicPopup CreateCustomPopup(string id, string label, Vector2 position, Vector2 size, bool devConfigurable = true)
        {
            var t = this.InstantiatePopup(UITransform);
            var rectT = Utils.GetRectTransform(t);
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicPopup>();
            uiDynamic.label = label;

            #if ENV_DEVELOPMENT
            {
                if(devConfigurable)
                {
                    AddRect(id, rectT);
                }
            }
            #endif

            return uiDynamic;
        }

        public UIDynamicPopup CreateCustomScrollablePopup(string label, Vector2 position, Vector2 size, bool devConfigurable = true) =>
            CreateCustomFilterablePopup(label, label, position, size, devConfigurable);

        public UIDynamicPopup CreateCustomScrollablePopup(string id, string label, Vector2 position, Vector2 size, bool devConfigurable = true)
        {
            var t = this.InstantiateScrollablePopup(UITransform);
            var rectT = Utils.GetRectTransform(t);
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicPopup>();
            uiDynamic.label = label;

            #if ENV_DEVELOPMENT
            {
                if(devConfigurable)
                {
                    AddRect(id, rectT);
                }
            }
            #endif

            return uiDynamic;
        }

        public UIDynamicPopup CreateCustomFilterablePopup(string label, Vector2 position, Vector2 size, bool devConfigurable = true) =>
            CreateCustomFilterablePopup(label, label, position, size, devConfigurable);

        public UIDynamicPopup CreateCustomFilterablePopup(string id, string label, Vector2 position, Vector2 size, bool devConfigurable = true)
        {
            var t = this.InstantiateFilterablePopup(UITransform);
            var rectT = Utils.GetRectTransform(t);
            rectT.anchoredPosition = position;
            rectT.sizeDelta = size;
            var uiDynamic = t.GetComponent<UIDynamicPopup>();
            uiDynamic.label = label;

            #if ENV_DEVELOPMENT
            {
                if(devConfigurable)
                {
                    AddRect(id, rectT);
                }
            }
            #endif

            return uiDynamic;
        }

        // ReSharper restore UnusedMember.Global
        // ReSharper restore MemberCanBePrivate.Global
        // ReSharper restore MemberCanBeProtected.Global

        // ReSharper disable once MemberCanBeMadeStatic.Global
        public void ConfigureCustomPopup(UIDynamicPopup uiDynamic, float height, bool upwards)
        {
            uiDynamic.popup.labelText.color = Color.black;

            if(height > 0f)
            {
                uiDynamic.popupPanelHeight = height;
            }

            if(upwards)
            {
                uiDynamic.popup.popupPanel.offsetMin += new Vector2(0, uiDynamic.popupPanelHeight + 60);
                uiDynamic.popup.popupPanel.offsetMax += new Vector2(0, uiDynamic.popupPanelHeight + 60);
            }
        }

        public void AddPrevNextButtons(UIDynamicPopup uiDynamic, bool cycle = false)
        {
            var popup = uiDynamic.popup;
            popup.labelText.alignment = TextAnchor.UpperCenter;
            popup.labelText.GetComponent<RectTransform>().anchorMax = new Vector2(0, 0.89f);

            /* Prev */
            {
                var t = this.InstantiateButton(uiDynamic.transform);
                var uiDynamicButton = t.GetComponent<UIDynamicButton>();
                uiDynamicButton.label = "<";
                if(cycle)
                {
                    uiDynamicButton.AddListener(() =>
                    {
                        popup.visible = false;
                        popup.SetPreviousOrLastValue();
                    });
                }
                else
                {
                    uiDynamicButton.AddListener(() =>
                    {
                        popup.visible = false;
                        popup.SetPreviousValue();
                    });
                }

                var rectT = t.GetComponent<RectTransform>();
                rectT.pivot = Vector2.zero;
                rectT.anchoredPosition = new Vector2(10, 0);
                rectT.sizeDelta = Vector2.zero;
                rectT.offsetMin = new Vector2(5, 5f);
                rectT.offsetMax = new Vector2(80, 70);
                rectT.anchorMin = Vector2.zero;
                rectT.anchorMax = Vector2.zero;
            }

            /* Next */
            {
                var t = this.InstantiateButton(uiDynamic.transform);
                var uiDynamicButton = t.GetComponent<UIDynamicButton>();
                uiDynamicButton.label = ">";
                if(cycle)
                {
                    uiDynamicButton.AddListener(() =>
                    {
                        popup.visible = false;
                        popup.SetNextOrFirstValue();
                    });
                }
                else
                {
                    uiDynamicButton.AddListener(() =>
                    {
                        popup.visible = false;
                        popup.SetNextValue();
                    });
                }

                var rectT = t.GetComponent<RectTransform>();
                rectT.pivot = Vector2.zero;
                rectT.anchoredPosition = Vector2.zero;
                rectT.sizeDelta = Vector2.zero;
                rectT.offsetMin = new Vector2(82, 5);
                rectT.offsetMax = new Vector2(157, 70);
                rectT.anchorMin = Vector2.zero;
                rectT.anchorMax = Vector2.zero;
            }
        }

        protected void OnDestroy()
        {
            DestroyImmediate(PluginUIEventsListener);
            PluginUIEventsListener = null;
        }

        #if ENV_DEVELOPMENT

        bool _devUICreated;
        const float SCALE_MULT = 100f;
        RectTransform _selectedRect;
        readonly Dictionary<string, RectTransform> _rectTransforms = new Dictionary<string, RectTransform>();
        readonly JSONStorableStringChooser _rectJssc = new JSONStorableStringChooser("Rect", new List<string>(), Constant.SELECT, "Rect");
        readonly JSONStorableFloat _posX = new JSONStorableFloat("Pos X", 0.00f, -20.00f, 20.00f);
        readonly JSONStorableFloat _posY = new JSONStorableFloat("Pos Y", 0.00f, -20.00f, 20.00f);
        readonly JSONStorableFloat _sizeX = new JSONStorableFloat("Size X", 0.00f, -15.00f, 15.00f);
        readonly JSONStorableFloat _sizeY = new JSONStorableFloat("Size Y", 0.00f, -15.00f, 15.00f);

        void AddRect(string name, RectTransform rectT) => _rectTransforms[$"{name}_{Guid.NewGuid().ToString().Substring(0, 4)}"] = rectT;

        void UpdateRectOptions()
        {
            var options = new List<string> { Constant.SELECT };
            options.AddRange(_rectTransforms.Keys);
            _rectJssc.choices = options;
        }

        [SuppressMessage("ReSharper", "UnusedMember.Global")]
        void CreateDevUI()
        {
            var rectPopup = CreateCustomFilterablePopup(_rectJssc.name, _rectJssc.name, new Vector2(1475, -1050), new Vector2(-555, 120), false);
            ConfigureCustomPopup(rectPopup, 800f, true);
            AddPrevNextButtons(rectPopup);
            _rectJssc.setCallbackFunction = SelectRect;
            _rectJssc.popup = rectPopup.popup;

            var syncButton = CreateCustomButton("Sync Options", new Vector2(2000, -1045), new Vector2(-921, 110), false);
            syncButton.AddListener(UpdateRectOptions);

            var posX = CreateCustomSlider(_posX.name, new Vector2(1475, -1200), false);
            posX.valueFormat = "F2";
            _posX.setCallbackFunction = value =>
            {
                if(_selectedRect)
                {
                    _selectedRect.anchoredPosition = SCALE_MULT * new Vector2(value, _posY.val);
                }
            };
            _posX.slider = posX.slider;

            var posY = CreateCustomSlider(_posY.name, new Vector2(1475, -1350), false);
            posY.valueFormat = "F2";
            _posY.setCallbackFunction = value =>
            {
                if(_selectedRect)
                {
                    _selectedRect.anchoredPosition = SCALE_MULT * new Vector2(_posX.val, value);
                }
            };
            _posY.slider = posY.slider;

            var sizeX = CreateCustomSlider(_sizeX.name, new Vector2(1475, -1500), false);
            sizeX.valueFormat = "F2";
            _sizeX.setCallbackFunction = value =>
            {
                if(_selectedRect)
                {
                    _selectedRect.sizeDelta = SCALE_MULT * new Vector2(value, _sizeY.val);
                }
            };
            _sizeX.slider = sizeX.slider;

            var sizeY = CreateCustomSlider(_sizeY.name, new Vector2(1475, -1650), false);
            sizeY.valueFormat = "F2";
            _sizeY.setCallbackFunction = value =>
            {
                if(_selectedRect)
                {
                    _selectedRect.sizeDelta = SCALE_MULT * new Vector2(_sizeX.val, value);
                }
            };
            _sizeY.slider = sizeY.slider;

            _devUICreated = true;
        }

        void SelectRect(string value)
        {
            _selectedRect = _rectTransforms[value];
            var anchoredPosition = _selectedRect.anchoredPosition;
            _posX.defaultVal = anchoredPosition.x / SCALE_MULT;
            _posY.defaultVal = anchoredPosition.y / SCALE_MULT;
            _posX.val = _posX.defaultVal;
            _posY.val = _posY.defaultVal;
            var sizeDelta = _selectedRect.sizeDelta;
            _sizeX.defaultVal = sizeDelta.x / SCALE_MULT;
            _sizeY.defaultVal = sizeDelta.y / SCALE_MULT;
            _sizeX.val = _sizeX.defaultVal;
            _sizeY.val = _sizeY.defaultVal;
        }

        #endif
    }
}
