//
using MeshVR;
using System;
using System.Collections;
using SimpleJSON;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using UnityEngine;
using UnityEngine.UI;

namespace everlaster
{
    sealed class SessionPluginsHook : ScriptBase
    {
        public List<string> BuiltInActionNames { get; } = new List<string>
        {
            "SaveToStore1",
            "SaveToStore2",
            "SaveToStore3",
            "RestoreAllFromStore1",
            "RestoreAllFromStore2",
            "RestoreAllFromStore3",
            "RestorePhysicsFromStore1",
            "RestorePhysicsFromStore2",
            "RestorePhysicsFromStore3",
            "RestoreAppearanceFromStore1",
            "RestoreAppearanceFromStore2",
            "RestoreAppearanceFromStore3",
            "RestoreAllFromDefaults",
            "RestorePhysicalFromDefaults",
            "RestoreAppearanceFromDefaults",
        };

        public override bool ShouldIgnore() => false;
        readonly List<SessionPlugin> _sessionPlugins = new List<SessionPlugin>();
        ChildrenChangedListener _sessionPluginsChangedListener;
        readonly static Regex _pluginNameRegex = Utils.NewRegex(@"^plugin#\d+_");

        bool _uiInitialized;
        UIDynamicTextField _pluginsListTextField;
        JSONStorableBool _applyParamsToSessionOnRestoreJsb;

        protected override void OnUIEnabled()
        {
            if(_uiInitialized)
            {
                return;
            }

            {
                var sb = new StringBuilder()
                    .AppendSize("\n", 20)
                    .AppendBold(nameof(SessionPluginsHook))
                    .AppendFormat("  <size=28>v{0}</size>", VERSION);
                var uiDynamic = CreateCustomTextField("Title", sb.ToString(), new Vector2(10, -104), new Vector2(-545, 90));
                uiDynamic.UItext.fontSize = 36;
                uiDynamic.backgroundColor = Color.clear;
            }

            {
                var uiDynamic = CreateCustomTextField("Plugins List", "", new Vector2(10, -900), new Vector2(-20, 800));
                uiDynamic.UItext.fontSize = 28;
                uiDynamic.UItext.horizontalOverflow = HorizontalWrapMode.Overflow;
                var scrollView = uiDynamic.transform.Find("Scroll View");
                var scrollRect = scrollView.GetComponent<ScrollRect>();
                scrollRect.horizontal = true;
                scrollRect.horizontalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;
                _pluginsListTextField = uiDynamic;
            }

            {
                var uiDynamic = CreateCustomToggle(_applyParamsToSessionOnRestoreJsb.name, new Vector2(10, -960), new Vector2(-530, 50));
                _applyParamsToSessionOnRestoreJsb.toggle = uiDynamic.toggle;
            }

            {
                var sb = new StringBuilder()
                    .Append("If enabled, all parameter values will be applied to session plugins when this")
                    .AppendLine(" plugin is restored (e.g. when loading a scene containing this plugin).")
                    .AppendLine()
                    .Append("If disabled, the saved parameter values are not used.");
                var uiDynamic = CreateCustomTextField("Info Text", sb.ToString(), new Vector2(10, -1220), new Vector2(-530, 250));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.fontSize = 28;
            }

            UpdatePluginsList();
            _uiInitialized = true;
        }

        public override void Init()
        {
            try
            {
                if(containingAtom.type == "SessionPluginManager")
                {
                    Loggr.Error("Add this plugin to an atom or to Scene Plugins", false);
                    return;
                }

                StartCoroutine(InitCo());
            }
            catch(Exception e)
            {
                Initialized = false;
                Loggr.Error($"Init: {e}");
            }
        }

        IEnumerator InitCo()
        {
            yield return new WaitForEndOfFrame();
            while(SuperController.singleton.isLoading)
            {
                yield return null;
            }

            if(Initialized == false)
            {
                yield break;
            }

            _applyParamsToSessionOnRestoreJsb = new JSONStorableBool("Apply Params To Session On Restore", true);
            RegisterBool(_applyParamsToSessionOnRestoreJsb);

            var atoms = GlobalSceneOptions.singleton.atomContainer.GetComponentsInChildren<Atom>();
            var sessionPluginManager = atoms.First(atom => atom.type == "SessionPluginManager");
            var pluginsListTransform = sessionPluginManager.transform.Find("SessionPluginManager/Plugins");
            yield return SyncSessionPluginsCo(pluginsListTransform);
            _sessionPluginsChangedListener = pluginsListTransform.gameObject.AddComponent<ChildrenChangedListener>();
            _sessionPluginsChangedListener.OnTransformChildrenChangedHandlers += SyncSessionPlugins;

            foreach(string paramName in BuiltInActionNames)
            {
                var action = GetAction(paramName);
                if(action != null)
                {
                    DeregisterAction(action);
                }
            }

            Initialized = true;
        }

        bool _inSyncSessionPluginsCo;
        int _delayFrames;

        void SyncSessionPlugins(Transform t)
        {
            if(_inSyncSessionPluginsCo)
            {
                _delayFrames++;
                return;
            }

            StartCoroutine(SyncSessionPluginsCo(t));
        }

        IEnumerator SyncSessionPluginsCo(Transform t)
        {
            /* Run coroutine only once in case of loading a session plugins preset containing multiple plugins,
             * since each added Transform in the plugins list triggers the callback.
             * Delays coroutine execution by 3 frames + 1 for each plugin in preset.
             */
            if(!_inSyncSessionPluginsCo)
            {
                yield return null;
                yield return null;
                yield return null;
            }

            _inSyncSessionPluginsCo = true;
            yield return WaitForFrames(_delayFrames);
            _delayFrames = 0;

            var retry = new List<Transform>();
            foreach(Transform container in t)
            {
                if(!container)
                {
                    continue;
                }

                var script = container.GetComponent<MVRScript>();
                if(!script)
                {
                    retry.Add(container);
                }
                else
                {
                    SetupPlugin(script);
                }
            }

            /* Wait until all Transforms actually contain the script component before proceeding to setup the storable params */
            while(retry.Count > 0)
            {
                const float timeout = 60;
                float deadline = Time.unscaledTime + timeout;

                yield return new WaitForSeconds(0.1f);

                for(int i = retry.Count - 1; i >= 0; i--)
                {
                    var container = retry[i];
                    var script = container.GetComponent<MVRScript>();
                    if(script)
                    {
                        SetupPlugin(script);
                        retry.RemoveAt(i);
                    }
                }

                if(Time.unscaledTime > deadline)
                {
                    Loggr.Error($"Unexpected error - some session plugin failed to initialize in {timeout:0}s", false);
                    yield break;
                }
            }

            CleanupRemoved();

            if(_uiInitialized)
            {
                UpdatePluginsList();
            }

            _inSyncSessionPluginsCo = false;
        }

        static IEnumerator WaitForFrames(int frames)
        {
            for(int i = 0; i < frames; i++)
            {
                yield return null;
            }
        }

        void CleanupRemoved()
        {
            for(int i = 0; i < _sessionPlugins.Count; i++)
            {
                var plugin = _sessionPlugins[i];
                if(plugin.IsRemoved)
                {
                    plugin.Cleanup();
                    _sessionPlugins.RemoveAt(i);
                }
            }
        }

        void SetupPlugin(MVRScript script)
        {
            string pluginName = _pluginNameRegex.Replace(script.name, "");
            if(_sessionPlugins.Exists(x => x.Name == pluginName))
            {
                return;
            }

            _sessionPlugins.Add(new SessionPlugin(pluginName, script, this));
        }

        readonly StringBuilder _pluginsListSb = new StringBuilder();

        void UpdatePluginsList()
        {
            _pluginsListSb.Clear().AppendSize("\n", 8);
            foreach(var plugin in _sessionPlugins)
            {
                _pluginsListSb.AppendLine(plugin.FullName);
            }

            _pluginsListTextField.text = _pluginsListSb.ToString();
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false) =>
            base.GetJSON(includePhysical, includeAppearance, true);

        public override void RestoreFromJSON(
            JSONClass jsonClass,
            bool restorePhysical = true,
            bool restoreAppearance = true,
            JSONArray presetAtoms = null,
            bool setMissingToDefault = true
        )
        {
            StartCoroutine(
                RestoreFromJSONCo(
                    jsonClass,
                    restorePhysical,
                    restoreAppearance,
                    presetAtoms,
                    setMissingToDefault
                )
            );
        }

        IEnumerator RestoreFromJSONCo(
            JSONClass jsonClass,
            bool restorePhysical,
            bool restoreAppearance,
            JSONArray presetAtoms,
            bool setMissingToDefault
        )
        {
            while(Initialized == null)
            {
                yield return null;
            }

            if(Initialized == false)
            {
                yield break;
            }

            _applyParamsToSessionOnRestoreJsb.RestoreFromJSON(jsonClass, restorePhysical, restoreAppearance, setMissingToDefault);
            if(_applyParamsToSessionOnRestoreJsb.val)
            {
                base.RestoreFromJSON(jsonClass, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
            }
        }

        new void OnDestroy()
        {
            try
            {
                base.OnDestroy();
                DestroyImmediate(_sessionPluginsChangedListener);
                _sessionPluginsChangedListener = null;

                foreach(var plugin in _sessionPlugins)
                {
                    plugin.Cleanup();
                }
            }
            catch(Exception e)
            {
                if(Initialized == true)
                {
                    SuperController.LogError($"OnDestroy: {e}");
                }
                else
                {
                    Debug.LogError($"OnDestroy: {e}");
                }
            }
        }
    }
}
